/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

/*
	UdmCORBA exception handling model

	description attribute gives	a detailed description of the exception
	code attribute gives an unique identification of the exception

	Code attributes:

		0 - Fatal exception: Fatal failure


		1 <-> 10 CORBA layer exceptions

			1 - Fatal exception: Caught a CORBA::Exception exception
			2 - Failed to narrow root POA
			3 - Failed to narrow root POA manager
			6 - Probably there is no active Backplane server.
			7 - Fatal error.


		11 <-> 20 Paradigm related exceptions

			11 - Paradigm is not found
			12 - A composition was not found in the paradigm
			13 - An association was not found in the paradigm
			14 - Paradigm exists
			15 - Paradigm is locked
			16 - Paradigm is in use
			17 - Paradigm file is invalid
			18 - Paradigm is unknown
			19 - A class was not found in the paradigm. Maybe the paradigms are different.
			20 - Paradigm is already unlocked


		21 <-> 40 UdmCORBA layer related exceptions

			21 - CORBAObject's setStringAttr method not implemented
			22 - CORBAObject's setBooleanAttr method not implemented
			23 - CORBAObject's setIntegerAttr method not implemented
			24 - CORBAObject's setRealAttr method not implemented
			25 - CORBAObject's setParent method not implemented
			26 - CORBAObject's detach method not implemented
			27 - CORBAObject's setChildren method not implemented
			28 - CORBAObject's createChild method not implemented
			29 - CORBAObject's setAssociation method not implemented
			30 - CORBAObject's __getdn method not implemented
			31 - Invalid session
			32 - Invalid document
			33 - fetchDocument error: Could not get the document
			34 - Subscribe failed: ToolAdaptor is not found
			35 - Failed to narrow to TranslatorFeed
			36 - Failed to narrow to TranslatorSink
			37 - publishDocument error: Could not publish the document. There is no active ToolAdaptor session.

		61 <-> 70 Manager related exceptions

			61 - Could not connect to the Backplane server
			62 - The password is incorrect
			63 - There is another active Manager application

	
		71 <-> 80 Translator related exceptions

			71 - Translator exists
			72 - Translator is in use
			73 - Invalid translator
			74 - Invalid translator arguments
			75 - Translator get_fromDN function is ambiguous, because two or more paradigms are available.
			76 - Translator get_fromDN function is ambiguous, because two or more paradigms with the same specified name are available.


		81 <-> 90 Tooladaptor related exceptions
			
			81 - Invalid ToolAdaptorType
			82 - ToolAdaptor createDatanetwork function is ambiguous, because two or more paradigms are available.
			83 - ToolAdaptor has to create a datanetwork first.
			85 - ToolAdaptor subscribe function is ambiguous, because two or more paradigms are available.
			86 - The %s v%s paradigm is not allowed for this type of ToolAdaptor.
			87 - Session of the ToolAdaptor could not be found.
			88 - ToolAdaptor could not fetch a remote paradigm because datanetwork is NULL.
			89 - CORBA layer was not properly initialized.

*/

#ifndef MOBIES_UDMCORBA_EXCEPTION_H
#define MOBIES_UDMCORBA_EXCEPTION_H

#include <exception>
#include <string>


class udmcorba_exception : public exception {

public:
	udmcorba_exception() throw();
	udmcorba_exception(const udmcorba_exception &e) throw() : description(e.description), exception_code(e.exception_code) {}
	udmcorba_exception(const string &d, const int c) throw() : description(d), exception_code(c) {}
	udmcorba_exception(const char *d, const int c) throw() : description(d), exception_code(c) {}

	const udmcorba_exception &operator =(const udmcorba_exception &e) throw() { 
		description = e.description; 
		exception_code = e.exception_code;
	}

	virtual ~udmcorba_exception() {}

	virtual const char *what() const throw() { 
		return description.c_str(); 
	}

	virtual const int code() const throw() {
		return exception_code;
	}

protected:
	int exception_code;
	string description;
};


class udmcorba_corbalayer_exception : public udmcorba_exception {
public:
	udmcorba_corbalayer_exception() throw();
	udmcorba_corbalayer_exception(const udmcorba_exception &e) throw() : udmcorba_exception(e) {}
	udmcorba_corbalayer_exception(const string &d, const int c) throw() : udmcorba_exception(d, c) {}
	udmcorba_corbalayer_exception(const char *d, const int c) throw() : udmcorba_exception(d, c) {}
};

class udmcorba_paradigm_exception : public udmcorba_exception {
public:
	udmcorba_paradigm_exception() throw();
	udmcorba_paradigm_exception(const udmcorba_exception &e) throw() : udmcorba_exception(e) {}
	udmcorba_paradigm_exception(const string &d, const int c) throw() : udmcorba_exception(d, c) {}
	udmcorba_paradigm_exception(const char *d, const int c) throw() : udmcorba_exception(d, c) {}
};

class udmcorba_manager_exception : public udmcorba_exception {
public:
	udmcorba_manager_exception() throw();
	udmcorba_manager_exception(const udmcorba_exception &e) throw() : udmcorba_exception(e) {}
	udmcorba_manager_exception(const string &d, const int c) throw() : udmcorba_exception(d, c) {}
	udmcorba_manager_exception(const char *d, const int c) throw() : udmcorba_exception(d, c) {}
};

class udmcorba_translator_exception : public udmcorba_exception {
public:
	udmcorba_translator_exception() throw();
	udmcorba_translator_exception(const udmcorba_exception &e) throw() : udmcorba_exception(e) {}
	udmcorba_translator_exception(const string &d, const int c) throw() : udmcorba_exception(d, c) {}
	udmcorba_translator_exception(const char *d, const int c) throw() : udmcorba_exception(d, c) {}
};

class udmcorba_tooladaptor_exception : public udmcorba_exception {
public:
	udmcorba_tooladaptor_exception() throw();
	udmcorba_tooladaptor_exception(const udmcorba_exception &e) throw() : udmcorba_exception(e) {}
	udmcorba_tooladaptor_exception(const string &d, const int c) throw() : udmcorba_exception(d, c) {}
	udmcorba_tooladaptor_exception(const char *d, const int c) throw() : udmcorba_exception(d, c) {}
};

#endif