#ifdef _WIN32
#include <UdmGme.h>			// these must be included to make SmartDataNetwork work
#endif
#include <UdmDom.h>
#include <Uml.h>
#include <UdmStatic.h>
#include <UmlExt.h>
#include <UdmBase.h>
#include <UdmInt.h>
#include <UdmCint.h>

#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include "config.h"

#ifndef __INTEGRATED_INTO_OTIF__
//#include "windview.h"
#endif

#include "iif_support.h"
#include "aif_support.h"
#include <algorithm>
#include <vector>
#include <list>


using namespace IIF;


//****************************************************************************************
// Gloabally accessed variables.
//****************************************************************************************

IIF::System iif_system;
Udm::SmartDataNetwork* iif;  
bool bFrameTimestamps   = FALSE;
bool bIgnoreFatalErrors = FALSE;
int output_timestamps = 0;

//****************************************************************************************
// Local common variables.
//****************************************************************************************

static int task_id = 0;

static const int MAX_SUBSCRIBE_PORTS = 4096;

static const unsigned __int64 max_longlong = 0xffffffffffffffff;

static COMPONENT_INSTANCE_LIST_ITEM component_instance_list[MAX_SUBSCRIBE_PORTS];
static int component_instance_list_i= 0;

static SUBSCRIBE_PORT_LIST_ITEM subscribe_port_list[MAX_SUBSCRIBE_PORTS];
static int subscribe_port_list_i= 0;

static ACTION_LIST_ITEM action_list[MAX_SUBSCRIBE_PORTS];
static int action_list_i= 0;

static Process process_list[MAX_SUBSCRIBE_PORTS];
static int process_list_i = 0;

static SUBSCRIBE_PORT_LIST_ITEM publish_port_list[MAX_SUBSCRIBE_PORTS];
static int publish_port_list_i= 0;

static SUBSCRIBE_PORT_LIST_ITEM framecontroller_subscribe_port_list[MAX_SUBSCRIBE_PORTS];
static int framecontroller_subscribe_port_list_i= 0;

static SUBSCRIBE_PORT_LIST_ITEM framecontroller_publish_port_list[MAX_SUBSCRIBE_PORTS];
static int framecontroller_publish_port_list_i= 0;

static unsigned __int64 largest_blocking_time = 0;

typedef enum THREAD_STATE {undefined=0, unknown, idle, active, preempted};
typedef enum OVERHEAD_CAUSE {overhead_unknown = 0, overhead_frame_start, overhead_frame_controller, overhead_previous_component};

typedef struct thread_list_item
{
	Thread        definition;
    THREAD_STATE state;
	 unsigned __int64 infrastructure_start_time;
	 OVERHEAD_CAUSE   infrastructure_overhead;
	 unsigned __int64 time_resumed;
	 unsigned __int64 unaccounted_for_time;
} THREAD_LIST_ITEM;

static THREAD_LIST_ITEM thread_list[MAX_SUBSCRIBE_PORTS];
static int thread_list_i = 0;

class CE_TIMESTAMP_BLOCK  
{
	 public:
		CE_TIMESTAMP_BLOCK() { }
	    ConsumedEventTimeStamp ce;
	    unsigned __int64 time;
};

static bool operator< (const class CE_TIMESTAMP_BLOCK &a, const class CE_TIMESTAMP_BLOCK &b)  
{ 
	if(a.time == b.time)
	{
		return ((b.ce.type() & 1) < (a.ce.type() & 1));
	}
	return (a.time < b.time); 
}

static vector<CE_TIMESTAMP_BLOCK>     ce_blocks;
static vector<SuppliedEventTimeStamp> se_timestamps; 
static vector<RemoteMethodTimeStamp>  rm_timestamps;
static vector<FrameTimeStamp>         fr_timestamps;
static vector<ModeTimeStamp>          md_timestamps;
static vector<PreemptionTimeStamp>    pe_timestamps;

static vector<CE_TIMESTAMP_BLOCK>::iterator     ce_block_i;
static vector<SuppliedEventTimeStamp>::iterator se_timestamp_i;
static vector<RemoteMethodTimeStamp>::iterator  rm_timestamp_i;
static vector<FrameTimeStamp>::iterator         fr_timestamp_i;
static vector<ModeTimeStamp>::iterator          md_timestamp_i;
static vector<PreemptionTimeStamp>::iterator    pe_timestamp_i;



//****************************************************************************************
// Local function prototypes.
//****************************************************************************************

int find_process_number(Process pr);
void collect_component_instances(void);
void collect_framecontrollers(void);
void collect_threads(void);
int find_matching_component_instance(IIF::System& system, IIF::SubscribePort& ce_subscribe_port );
int find_matching_component_instance(IIF::System& system, IIF::PublishPort& se_publish_port );
int find_matching_framecontroller(IIF::System& system, IIF::SubscribePort& ce_subscribe_port );
int find_matching_framecontroller(IIF::System& system, IIF::PublishPort& se_publish_port );
int find_thread( Thread& nt );
int find_rate( Rate& rt );
BOOL get_next_timestamp_info(TIMESTAMP_INFO &TimestampInfo, bool advance);
TIMESTAMP_TYPE next_timestamp_type();
unsigned __int64 next_ce_event_timestamp();
unsigned __int64 next_se_event_timestamp();
unsigned __int64 next_pe_event_timestamp();
unsigned __int64 next_rm_event_timestamp();
unsigned __int64 next_fr_event_timestamp();
unsigned __int64 next_md_event_timestamp();
 
bool se_comp_timestamps( SuppliedEventTimeStamp a, SuppliedEventTimeStamp b);
bool rm_comp_timestamps( RemoteMethodTimeStamp  a, RemoteMethodTimeStamp  b);
bool fr_comp_timestamps( FrameTimeStamp         a, FrameTimeStamp         b);
bool md_comp_timestamps( ModeTimeStamp          a, ModeTimeStamp          b);
bool pe_comp_timestamps( PreemptionTimeStamp    a, PreemptionTimeStamp    b);


//****************************************************************************************
//                            Local function definitions.
//****************************************************************************************

//****************************************************************************************
// Function navigates the IIF data tree and builds a list of the component instances.
//****************************************************************************************
void collect_component_instances(void)
{
  int process_number = -1;

	set<Processor> zprocessors = iif_system.Processor_kind_children();
	set<Processor>::iterator zprocessors_i = zprocessors.begin();
	for(int xx=0;zprocessors_i != zprocessors.end();xx++)
	{
		Processor zp = *zprocessors_i;
		string zpname = zp.name();

		set<Process> processes = zp.Process_kind_children();
		set<Process>::iterator processes_i = processes.begin();
		for(int xx=0;processes_i != processes.end();xx++)
		{
			Process pr = *processes_i;
			string prname = pr.name();

            process_list[process_list_i++] = pr;
			process_number++;

			set<Component> components = pr.Component_kind_children();
			set<Component>::iterator components_i = components.begin();
			for(int xx=0;components_i != components.end();xx++)
			{
				Component cp  = *components_i;
				string cpname = cp.name();
				__int64 grpid     = cp.GroupId();
				__int64 itemid    = cp.ItemId();
				string sgrpid     = __i2s(grpid);
				string sitemid    = __i2s(itemid);

                component_instance_list[component_instance_list_i].cp    = cp;
                component_instance_list[component_instance_list_i].group = grpid;
                component_instance_list[component_instance_list_i].item  = itemid;
 				component_instance_list[component_instance_list_i].current_mode = 0;
				component_instance_list[component_instance_list_i].mode_name[0]="default";
                component_instance_list_i++;

				set<SubscribePort> subscribe_ports = cp.SubscribePort_kind_children();
				set<SubscribePort>::iterator subscribe_ports_i = subscribe_ports.begin();
				for(int xx=0;subscribe_ports_i != subscribe_ports.end();xx++)
				{
					SubscribePort sp = *subscribe_ports_i;
					string spname = sp.name();

#ifndef __INTEGRATED_INTO_OTIF__
					outStream << "Processor " << zpname << "\n Process " << prname << "\n Component "
							<< cpname  << "\n  GroupId="<<  grpid << "  ItemId=" << itemid
							<< "\n  Subscribe Port: " << xx << " " << spname << endl;
#endif
					subscribe_port_list[subscribe_port_list_i].sp    = sp;
					subscribe_port_list[subscribe_port_list_i].cp    = cp;
					subscribe_port_list[subscribe_port_list_i].pr    = pr;
					subscribe_port_list[subscribe_port_list_i].item  = cp.ItemId();
					subscribe_port_list[subscribe_port_list_i].group = cp.GroupId();
					subscribe_port_list[subscribe_port_list_i].task_id = task_id;
					subscribe_port_list_i++;
					subscribe_ports_i++;
					string sps =  cpname + "<" + sgrpid + ","  + sitemid + ">"  + spname;
#ifndef __INTEGRATED_INTO_OTIF__
//					windview_define_task( process_number, (char*)sps.c_str(),task_id++);
#endif
				}

				set<PublishPort> publish_ports = cp.PublishPort_kind_children();
				set<PublishPort>::iterator publish_ports_i = publish_ports.begin();
				for(int pxx=0;publish_ports_i != publish_ports.end();pxx++)
				{
					PublishPort pp = *publish_ports_i;
					string ppname = pp.name();

#ifndef __INTEGRATED_INTO_OTIF__
					outStream << "Processor " << zpname << "\n Process " << prname << "\n Component "
							<< cpname  << "\n  GroupId="<<  grpid << "  ItemId=" << itemid
							<< "\n  Publis Port: " << pxx << " " << ppname << endl;
#endif
					publish_port_list[publish_port_list_i].pp    = pp;
					publish_port_list[publish_port_list_i].cp    = cp;
					publish_port_list[publish_port_list_i].pr    = pr;
					publish_port_list[publish_port_list_i].item  = cp.ItemId();
					publish_port_list[publish_port_list_i].group = cp.GroupId();
					publish_port_list[publish_port_list_i].task_id = task_id;
					publish_port_list_i++;
					publish_ports_i++;
					string pps = "pp_" + ppname;
				}

				set<Action> actions = cp.Action_kind_children();
				set<Action>::iterator action_i = actions.begin();
				for(int axx=0;action_i != actions.end();axx++)
				{
					Action a = *action_i;
					string aname = a.name();

#ifndef __INTEGRATED_INTO_OTIF__
					outStream << "Processor " << zpname << "\n Process " << prname << "\n Component "
							<< cpname  << "\n  GroupId="<<  grpid << "  ItemId=" << itemid
							<< "\n  Action: " << axx << " " << aname << endl;
#endif
					action_list[action_list_i].a     =   a;
					action_list[action_list_i].cp      = cp;
					action_list[action_list_i].pr      = pr;
					action_list[action_list_i].item    = cp.ItemId();
					action_list[action_list_i].group   = cp.GroupId();
					action_list[action_list_i].task_id = task_id;

					action_list_i++;
					action_i++;
				}

				
				components_i++;
			}
			processes_i++;
		}
		zprocessors_i++;
	}

	return;
}

//****************************************************************************************
// Function navigates the IIF data tree and builds a list of the frame controllers.
//****************************************************************************************
void collect_framecontrollers(void)
{
	set<Processor> zprocessors = iif_system.Processor_kind_children();
	set<Processor>::iterator zprocessors_i = zprocessors.begin();
	for(int xx=0;zprocessors_i != zprocessors.end();xx++)
	{
		Processor zp = *zprocessors_i;
		string zpname = zp.name();

		set<Process> processes = zp.Process_kind_children();
		set<Process>::iterator processes_i = processes.begin();
		for(int xx=0;processes_i != processes.end();xx++)
		{
			Process pr = *processes_i;
			string prname = pr.name();

			set<FrameController> framecontrollers =  pr.FrameController_kind_children();
			set<FrameController>::iterator framecontroller_i = framecontrollers.begin();
			for(int xx=0;framecontroller_i != framecontrollers.end();xx++)
			{
				FrameController fc  = *framecontroller_i;
				string fcname = fc.name();

				set<SubscribePort> framecontroller_subscribe_ports = fc.SubscribePort_kind_children();
				set<SubscribePort>::iterator framecontroller_subscribe_ports_i = framecontroller_subscribe_ports.begin();
				for(int xx=0; framecontroller_subscribe_ports_i != framecontroller_subscribe_ports.end();xx++)
				{
					SubscribePort sp = *framecontroller_subscribe_ports_i;
					string spname = sp.name();

#ifndef __INTEGRATED_INTO_OTIF__
					outStream << "Processor " << zpname << "\n Process " << prname << "\n FrameController "
							<< fcname  << "\n  Subscribe Port: " << xx << " " << spname << endl;
#endif
					framecontroller_subscribe_port_list[framecontroller_subscribe_port_list_i].fc = fc;
					framecontroller_subscribe_port_list[framecontroller_subscribe_port_list_i].sp = sp;

					framecontroller_subscribe_port_list[framecontroller_subscribe_port_list_i].task_id    = task_id;
					string sps =  fcname + "<" + ">"  + spname;

#ifndef __INTEGRATED_INTO_OTIF__
					if(bFrameTimestamps)
					{
//					   windview_define_task( find_process_number(pr), (char*)sps.c_str(),task_id++);
					}
#endif

					framecontroller_subscribe_port_list_i++;
					framecontroller_subscribe_ports_i++;
				}

				set<PublishPort> framecontroller_publish_ports = fc.PublishPort_kind_children();
				set<PublishPort>::iterator framecontroller_publish_ports_i = framecontroller_publish_ports.begin();
				for(int pxx=0;framecontroller_publish_ports_i != framecontroller_publish_ports.end();pxx++)
				{
					PublishPort pp = *framecontroller_publish_ports_i;
					string ppname = pp.name();

#ifndef __INTEGRATED_INTO_OTIF__
					outStream << "Processor " << zpname << "\n Process " << prname << "\n FrameController "
							<< fcname << "\n  Publish Port: " << pxx << " " << ppname << endl;
#endif
					framecontroller_publish_port_list[framecontroller_publish_port_list_i].pp    = pp;
					framecontroller_publish_port_list[framecontroller_publish_port_list_i].fc    = fc;
					framecontroller_publish_port_list_i++;
					framecontroller_publish_ports_i++;
				}
				framecontroller_i++;
			}
			processes_i++;
		}
		zprocessors_i++;
	}

	return;
}

//****************************************************************************************
// Function navigates the IIF data tree and builds a list of the threads.
//****************************************************************************************
void collect_threads(void)
{
	set<Processor> zprocessors = iif_system.Processor_kind_children();
	set<Processor>::iterator zprocessors_i = zprocessors.begin();
	for(int xx=0;zprocessors_i != zprocessors.end();xx++)
	{
		Processor zp = *zprocessors_i;
		string zpname = zp.name();

		set<Process> processes = zp.Process_kind_children();
		set<Process>::iterator processes_i = processes.begin();
		for(int xx=0;processes_i != processes.end();xx++)
		{
			Process pr = *processes_i;
			string prname = pr.name();

			set<Thread> Threads = pr.Thread_kind_children();
			set<Thread>::iterator Threads_i = Threads.begin();
			for(int xx=0;Threads_i != Threads.end();xx++)
			{
				Thread th  = *Threads_i;
				string thname = th.name();
				set<Rate> rates = th.Rate_kind_children();
				set<Rate>::iterator rates_i = rates.begin();

				
#ifndef __INTEGRATED_INTO_OTIF__
				outStream << "Processor " << zpname << "\n Process " << prname << "\n Thread "<< thname << " priority "  << th.threadPriority() << " "; 
				outStream << "rates [ ";
				for(;rates_i != rates.end(); rates_i++)
				{
					Rate rate = *rates_i;
					string ratename = rate.name();

					outStream << ratename << " ";
				}
				outStream << "]" << endl;
#endif

				thread_list[thread_list_i].state = unknown;
				thread_list[thread_list_i].time_resumed         = 0;
				thread_list[thread_list_i].unaccounted_for_time = 0;
				thread_list[thread_list_i].definition = th;
				thread_list[thread_list_i].infrastructure_start_time = 0;
				thread_list[thread_list_i].infrastructure_overhead   = overhead_unknown;
				thread_list_i++;
				Threads_i++;
			}
			processes_i++;
		}
		zprocessors_i++;
	}

	return;
}

//****************************************************************************************
// Function finds the index of the subscribe port for the matching component instance.
//****************************************************************************************
int find_matching_component_instance(IIF::System& system, IIF::SubscribePort& ce_subscribe_port )
{
	int i;

	for (i=0;i<subscribe_port_list_i;i++)
	{
		if(subscribe_port_list[i].sp == ce_subscribe_port)
		{
			return i;
		}
	}
	return -1;
}


string iif_get_component_mode_name( __int64 groupid, __int64 itemid, int mode_index )
{
  int i;

  for(i=0;i<component_instance_list_i;i++)
  {
	if(component_instance_list[i].group != groupid) continue;
	if(component_instance_list[i].item  != itemid)  continue;
	return component_instance_list[i].mode_name[mode_index];
  }
  return "";
}


void iif_show_component_instance_list(void)
{
  int i,z;
  string mn;

  outStream << "Component Instance Mode Details" << endl;

  for(i=0;i<component_instance_list_i;i++)
  {
     Component cp = component_instance_list[i].cp;
	 string cn = cp.name();
	 __int64 i = cp.ItemId();
	 __int64 g = cp.GroupId();

     outStream << "  " << cn << "<" << __i2s(g) << "," << __i2s(i) << "> : { ";
	 for(z=0;;z++)
	 {
		mn=iif_get_component_mode_name(g,i,z);
		if(mn=="") break;
	    outStream << mn << " ";
	 }
	 outStream << "}" << endl;
  }
}


int find_process_number(Process pr)
{
	int i;

	for(i=0;i<process_list_i;i++)
	{
		if(pr == process_list[i]) return i;
	}

    outStream << "Error: Timestamp with no process  " << endl;
    cerr      << "Error: Timestamp with no process  " << endl;
    if(!bIgnoreFatalErrors)
	{
  	  _sleep(3000);
	  _exit(__LINE__);
	}		
  return 0;
}


int get_component_current_mode( __int64 groupid, __int64 itemid)
{
  int i;

  for(i=0;i<component_instance_list_i;i++)
  {
	if(component_instance_list[i].group != groupid) continue;
	if(component_instance_list[i].item  != itemid)  continue;
	return component_instance_list[i].current_mode;
  }

  outStream << "Error: Mode lookup failed  " << endl;
  cerr      << "Error: Mode lookup failed  " << endl;
  if(!bIgnoreFatalErrors)
  {
	_sleep(3000);
	_exit(__LINE__);
  }		
 return 0;
}

//****************************************************************************************
// Function finds the index of the subscribe port for the matching component instance.
//****************************************************************************************
int find_matching_component_instance(IIF::System& system, int groupid, int itemid )
{
	int i;

	for (i=0;i<subscribe_port_list_i;i++)
	{
		if((subscribe_port_list[i].group == groupid) && (subscribe_port_list[i].item == itemid))
		{
			return i;
		}
	}
	return -1;
}


//****************************************************************************************
// Function finds the index of the subscribe port for the matching component instance.
//****************************************************************************************
int find_matching_component_action_list_instance(IIF::System& system, int groupid, int itemid )
{
	int i;

	for (i=0;i<action_list_i;i++)
	{
		if((action_list[i].group == groupid) && (action_list[i].item == itemid))
		{
			return i;
		}
	}
	return -1;
}


//****************************************************************************************
// Function finds the index of the subscribe port for the matching component instance.
//****************************************************************************************
int find_matching_in_component_instance_list(__int64 groupid, __int64 itemid )
{
	int i;

	for (i=0;i<component_instance_list_i;i++)
	{
		if((component_instance_list[i].group == groupid) && (component_instance_list[i].item == itemid))
		{
			return i;
		}
	}
	return -1;
}



//****************************************************************************************
//****************************************************************************************
int find_or_add_mode(int idx, string this_mode)
{
 int i;

  for(i=0;i<MAX_MODES;i++)
  {
	  if(component_instance_list[idx].mode_name[i] == this_mode)
	  {
		  return i;
	  }
	  if(component_instance_list[idx].mode_name[i]=="")
	  {
		  component_instance_list[idx].mode_name[i] = this_mode;
		  return i;
	  }
  }

  outStream << "Error: Out of modes to add  " << endl;
  cerr      << "Error: Out of modes to add  " << endl;
  if(!bIgnoreFatalErrors)
  {
	_sleep(3000);
	_exit(__LINE__);
  }		
 return 0;
}

//****************************************************************************************
//****************************************************************************************
string update_component_instance_modes(IIF::System& system, IIF::ModeTimeStamp mts)
{
 string smodes = "";
 set<ComponentMode> modes = mts.ComponentMode_kind_children();
 set<ComponentMode>::iterator mode_i;
 int i;

 for(mode_i = modes.begin();mode_i != modes.end();mode_i++)
 {
	ComponentMode this_mode = *mode_i;
	string sthis_mode;

	sthis_mode = this_mode.mode();

	if(sthis_mode == "")
    {
	  outStream << "Error: Bad mode  " << endl;
	  cerr      << "Error: Bad mode  " << endl;
	  if(!bIgnoreFatalErrors)
	  {
		_sleep(3000);
		_exit(__LINE__);
	  }		
	  continue;
	}


	smodes += sthis_mode;
	smodes += "{";

	Component cp = this_mode.Component();
	__int64 grpid      = cp.GroupId();
	__int64 itemid     = cp.ItemId();
	string sgrpid  = __i2s(grpid);
	string sitemid = __i2s(itemid);

	smodes +=cp.name();
	smodes += "<";
	smodes += sgrpid;
	smodes += ",";
	smodes += sitemid;
	smodes += ">} ";

    i = find_matching_in_component_instance_list( grpid, itemid );
	if(i >= 0)
	{
	  component_instance_list[i].current_mode = find_or_add_mode(i, sthis_mode);
	  continue;
	}

    outStream << "Error: Component instance with matching mode not found <" << __i2s(grpid) << "," << __i2s(itemid) << ">" << endl;
    cerr      << "Error: Component instance with matching mode not found <" << __i2s(grpid) << "," << __i2s(itemid) << ">" << endl;
	if(!bIgnoreFatalErrors)
	{
	  _sleep(3000);
	  _exit(__LINE__);
	}		
	continue;
 }

 return smodes;
}

//****************************************************************************************
// Function finds the index of the publish port for the matching component instance.
//****************************************************************************************
int find_matching_component_instance(IIF::System& system, IIF::PublishPort& se_publish_port )
{
	int i;

	for (i=0;i<publish_port_list_i;i++)
	{
		if(publish_port_list[i].pp == se_publish_port)
		{
			return i;
		}
	}
	return -1;
}


//****************************************************************************************
// Function finds the index of the subscribe port for the matching frame controller.
//****************************************************************************************
int find_matching_framecontroller(IIF::System& system, IIF::SubscribePort& ce_subscribe_port )
{
	int i;

	for (i=0;i<framecontroller_subscribe_port_list_i;i++)
	{
		if(framecontroller_subscribe_port_list[i].sp == ce_subscribe_port)
		{
			return i;
		}
	}
	return -1;
}

//****************************************************************************************
// Function finds the index of the publish port for the matching frame controller.
//****************************************************************************************
int find_matching_framecontroller(IIF::System& system, IIF::PublishPort& se_publish_port )
{
	int i;

	for (i=0;i<framecontroller_publish_port_list_i;i++)
	{
		if(framecontroller_publish_port_list[i].pp == se_publish_port)
		{
			return i;
		}
	}
	return -1;
}

//****************************************************************************************
// Function outputs the index of the thread for the matching thread.
//****************************************************************************************
int find_thread( Thread& nt )
{
	int i;

	for(i=0;i<thread_list_i;i++)
	{
		if(nt == thread_list[i].definition) 
		{
			return i;
		}
	}
	return -1;
}

//****************************************************************************************
// Function outputs the index of the action for the matching action.
//****************************************************************************************
int find_action( RemoteMethodTimeStamp& rm )
{
	int i;
	Action a = rm.Action();

	for(i=0;i<action_list_i;i++)
	{
		if(a == action_list[i].a) 
		{
			return i;
		}
	}
	return -1;
}


//****************************************************************************************
// Function outputs the index of the thread for the matching rate
//****************************************************************************************
int find_rate( Rate& rt )
{
	int i;
	int r = -1;

	for(i=0;i<thread_list_i;i++)
	{
	  set<Rate> zRates = thread_list[i].definition.Rate_kind_children();
	  set<Rate>::iterator zRates_i = zRates.begin();
	  for(int xx=0;zRates_i != zRates.end();xx++)
	  {
	    Rate this_rate = *zRates_i;
		string ratename = this_rate.name();
		
		if(rt == this_rate) 
		{
		  if(r < 0)
			{
			  r = i;
			}
		   else
			{
			  outStream << "Error: Too many threads with rate " << ratename << endl;
			  cerr      << "Error: Too many threads with rate " << ratename <<  endl;
			  if(!bIgnoreFatalErrors)
			  {
			    _sleep(3000);
			    _exit(__LINE__);
			  }
			}
		}
		zRates_i++;
	  }
	}
	return r;
}

//****************************************************************************************
// Function gets the timestamp information for the next timestamp in the list.
//****************************************************************************************
BOOL get_next_timestamp_info(TIMESTAMP_INFO &TimestampInfo, bool advance)
{
	TIMESTAMP_TYPE c;
	int x;

	c = next_timestamp_type();
	TimestampInfo.iIndex = -1;
    TimestampInfo.bFrameController = FALSE;

	switch(c)
	{
		case eConsumedEvent:
		{
			ConsumedEventTimeStamp ce_timestamp = ce_block_i->ce;
			Thread ce_thread = ce_timestamp.thread();
			Process ce_process = ce_thread.Process_parent();

			SubscribePort ce_subscribe_port = ce_timestamp.SubscribePort();

			TimestampInfo.pr = ce_process;
			TimestampInfo.strType = CE_TIMESTAMP;
			
			TimestampInfo.llTimestamp = ce_block_i->time;


			TimestampInfo.strAction = (ce_timestamp.type() & 1) ? END_EVENT : BEGIN_EVENT;
			TimestampInfo.strThreadID = ce_thread.name();

			if ((x = find_matching_component_instance( iif_system, ce_subscribe_port)) >= 0) {
				TimestampInfo.strSubscribePort = subscribe_port_list[x].sp.name();
			    TimestampInfo.bFrameController = FALSE;
				TimestampInfo.iIndex = x;
			}
			else if ((x = find_matching_framecontroller( iif_system, ce_subscribe_port)) >= 0) {
				TimestampInfo.strSubscribePort = framecontroller_subscribe_port_list[x].sp.name();
			    TimestampInfo.bFrameController = TRUE;
				TimestampInfo.iIndex = x;
			}
			else {
				TimestampInfo.strSubscribePort = NO_DATA_FIELD;
			}

			if(advance)	ce_block_i++;
		}
		break;

		case eSuppliedEvent:
		{
			SuppliedEventTimeStamp se_timestamp = *se_timestamp_i;
			Thread se_thread = se_timestamp.thread();
			Process se_process = se_thread.Process_parent();
			PublishPort se_publish_port = se_timestamp.PublishPort();

			TimestampInfo.pr = se_process;
			TimestampInfo.strType = SE_TIMESTAMP;
			TimestampInfo.llTimestamp = __s2u(se_timestamp.timestamp());
			TimestampInfo.strAction = (se_timestamp.type() & 1) ? END_EVENT : BEGIN_EVENT;
			TimestampInfo.strThreadID = NO_DATA_FIELD;

			if ((x = find_matching_component_instance( iif_system, se_publish_port)) >= 0) {
				TimestampInfo.strSubscribePort = publish_port_list[x].pp.name();
			    TimestampInfo.bFrameController = FALSE;
				TimestampInfo.iIndex = x;
			}
			else if ((x = find_matching_framecontroller( iif_system, se_publish_port)) >= 0) {
				TimestampInfo.strSubscribePort = framecontroller_publish_port_list[x].pp.name();
			    TimestampInfo.bFrameController = TRUE;
				TimestampInfo.iIndex = x;
			}
			else {
				TimestampInfo.strSubscribePort = NO_DATA_FIELD;
			}

			if(advance)	se_timestamp_i++;
		}
		break;

		case ePreemption:
		{
			PreemptionTimeStamp pe_timestamp = *pe_timestamp_i;
			Thread th = pe_timestamp.Thread();

			Process pe_process = th.Process_parent();

			TimestampInfo.pr = pe_process;
			TimestampInfo.strType = PE_TIMESTAMP;
			TimestampInfo.llTimestamp = __s2u(pe_timestamp.timestamp());
			TimestampInfo.strAction = (pe_timestamp.type()) ? RETURN_TO_THREAD : LEAVE_THREAD;
			TimestampInfo.strSubscribePort = NO_DATA_FIELD;

			if ((x = find_thread(th)) >= 0) {
				TimestampInfo.strThreadID = thread_list[x].definition.name();
			} else {
				TimestampInfo.strThreadID = NO_DATA_FIELD;
			}

			if(advance)	pe_timestamp_i++;
		}
		break;

		case eFrame:
		{
			FrameTimeStamp this_fr_timestamp = *fr_timestamp_i;

            TimestampInfo.pr = 0;  // n/a
			TimestampInfo.strType = FR_TIMESTAMP;
			TimestampInfo.llTimestamp = __s2u(this_fr_timestamp.timestamp());
			TimestampInfo.strAction = (this_fr_timestamp.type()) ? END_EVENT : BEGIN_EVENT;
			TimestampInfo.strSubscribePort = NO_DATA_FIELD;

			Rate rate = this_fr_timestamp.Rate();

			if ((x = find_rate(rate)) >= 0) {
				TimestampInfo.strThreadID = thread_list[x].definition.name();
			} else {
				TimestampInfo.strThreadID = NO_DATA_FIELD;
			}

			if(advance)	fr_timestamp_i++;
		}
		break;

		case eRemoteMethod:
		{
			RemoteMethodTimeStamp this_rm_timestamp = *rm_timestamp_i;
			Thread th = this_rm_timestamp.thread();
			Process rm_process = th.Process_parent();
			int x = find_thread(th);

            TimestampInfo.pr = rm_process; 
			TimestampInfo.strType = RM_TIMESTAMP;
			TimestampInfo.llTimestamp = __s2u(this_rm_timestamp.timestamp());
			TimestampInfo.strSubscribePort = NO_DATA_FIELD;
			switch(this_rm_timestamp.type() & 3)
			{
			case 2:
			case 3:
				   TimestampInfo.strAction = NO_DATA_FIELD;
				   break;
			case 0:
				   TimestampInfo.strAction = BEGIN_RM;
				   break;
			case 1:
				   TimestampInfo.strAction = END_RM;
				   break;
			}

            if(x<0)
			  TimestampInfo.strThreadID = NO_DATA_FIELD;
			else
			  TimestampInfo.strThreadID = thread_list[x].definition.name();

			TimestampInfo.rm     = this_rm_timestamp;
			TimestampInfo.action = this_rm_timestamp.Action();

			if(advance)	rm_timestamp_i++;
		}
		break;

		case eMode:
		{
			ModeTimeStamp this_md_timestamp = *md_timestamp_i;

            TimestampInfo.pr = 0;  // n/a
			TimestampInfo.strType = MD_TIMESTAMP;
			TimestampInfo.llTimestamp = __s2u(this_md_timestamp.timestamp());
			TimestampInfo.strAction = update_component_instance_modes( iif_system, this_md_timestamp );
			TimestampInfo.strSubscribePort = NO_DATA_FIELD;

		    TimestampInfo.strThreadID = NO_DATA_FIELD;

			if(advance)	md_timestamp_i++;
		}
		break;

		case eNone:
		{
			return(FALSE);
		}
		break;

	} // end switch()

	return(TRUE);
}

//****************************************************************************************
// Functions determines the next timestamp type; used by the processing procedure.
//****************************************************************************************
TIMESTAMP_TYPE next_timestamp_type()
{
  unsigned __int64 rv,tv;
  TIMESTAMP_TYPE r=eNone;

  rv = max_longlong;

  if ((tv=next_pe_event_timestamp()) < rv) {rv = next_pe_event_timestamp(); r = ePreemption;}
  if ((tv=next_md_event_timestamp()) < rv) {rv = next_md_event_timestamp(); r = eMode;}
  if ((tv=next_fr_event_timestamp()) < rv) {rv = next_fr_event_timestamp(); r = eFrame;}
  if ((tv=next_rm_event_timestamp()) < rv) {rv = next_rm_event_timestamp(); r = eRemoteMethod;}
  if ((tv=next_se_event_timestamp()) < rv) {rv = next_se_event_timestamp(); r = eSuppliedEvent;}
  if ((tv=next_ce_event_timestamp()) < rv) {rv = next_ce_event_timestamp(); r = eConsumedEvent;}
  return r;
}

//****************************************************************************************
// Functions for returning the next timestamp in a set; used by the processing procedure.
//****************************************************************************************

unsigned __int64 next_ce_event_timestamp(){	if(ce_block_i == ce_blocks.end()) return max_longlong; return ce_block_i->time;} // could use template
unsigned __int64 next_se_event_timestamp(){	if(se_timestamp_i == se_timestamps.end()) return max_longlong; return __s2u(se_timestamp_i->timestamp());} // could  use template
unsigned __int64 next_pe_event_timestamp(){	if(pe_timestamp_i == pe_timestamps.end()) return max_longlong; return __s2u(pe_timestamp_i->timestamp());} // could  use template
unsigned __int64 next_rm_event_timestamp(){	if(rm_timestamp_i == rm_timestamps.end()) return max_longlong; return __s2u(rm_timestamp_i->timestamp());} // could  use template
unsigned __int64 next_fr_event_timestamp(){	if(fr_timestamp_i == fr_timestamps.end()) return max_longlong; return __s2u(fr_timestamp_i->timestamp());} // could  use template
unsigned __int64 next_md_event_timestamp(){	if(md_timestamp_i == md_timestamps.end()) return max_longlong; return __s2u(md_timestamp_i->timestamp());} // could  use template

//****************************************************************************************
// Functions for comparison of timestamps; used by sorting procedure.
//****************************************************************************************

bool se_comp_timestamps( SuppliedEventTimeStamp a, SuppliedEventTimeStamp b) { return  __s2u(a.timestamp()) <  __s2u(b.timestamp()); } // could  use template
bool rm_comp_timestamps( RemoteMethodTimeStamp  a, RemoteMethodTimeStamp  b) { return  __s2u(a.timestamp()) <  __s2u(b.timestamp()); } // could  use template
bool fr_comp_timestamps( FrameTimeStamp         a, FrameTimeStamp         b) { return  __s2u(a.timestamp()) <  __s2u(b.timestamp()); } // could  use template
bool md_comp_timestamps( ModeTimeStamp          a, ModeTimeStamp          b) { return  __s2u(a.timestamp()) <  __s2u(b.timestamp()); } // could  use template
bool pe_comp_timestamps( PreemptionTimeStamp    a, PreemptionTimeStamp    b) 
{ 
	// If timestamps match.
	if( __s2u(a.timestamp()) == __s2u(b.timestamp()))
	{
		// Make sure thread leaves before it returns.
		return b.type() > a.type();
	}
	return  __s2u(a.timestamp()) <  __s2u(b.timestamp()); 
}


//****************************************************************************************
//                            Global function definitions.
//****************************************************************************************


//****************************************************************************************
// Function closes the IIF data object.
//****************************************************************************************
void iif_close(void)
{
	iif->CloseNoUpdate();
}

//****************************************************************************************
// Function computes the WCET for each component subscribe port.
//****************************************************************************************
int zloops = 0;  // For debug only.
static bool first_time = TRUE;

void iif_compute_wcet(void)
{
	int i,z;
	string strCompSubscribePort;
	TIMESTAMP_INFO timestampInfo;

	zloops = 0;

#ifndef __INTEGRATED_INTO_OTIF__
	outStream << endl;
#endif

	//************** Initialize the subscribe port and action lists.*******************
	for (i=0; i<subscribe_port_list_i; i++) 
	{
		subscribe_port_list[i].strThreadID = "";
		subscribe_port_list[i].strPreemptor = "";
		subscribe_port_list[i].nCurState = Idle;
		subscribe_port_list[i].llTimeStarted = 0;
		subscribe_port_list[i].llTimeWentActive = 0;
		subscribe_port_list[i].llExecutionTime = 0;
		subscribe_port_list[i].WCBT            = 0;
		subscribe_port_list[i].llWallClockTime = 0;
		for(z=0;z<MAX_MODES;z++)
		{
		  subscribe_port_list[i].llWCExecutionTime[z] = 0;
		  subscribe_port_list[i].llWCWallClockTime[z] = 0;
		}
		subscribe_port_list[i].recursion_level = 0;
	}

	for (i=0; i<action_list_i; i++) 
	{
		action_list[i].strThreadID      = "";
		action_list[i].strPreemptor     = "";
		action_list[i].nCurState        = Idle;
		action_list[i].time_consumed    = 0;
		action_list[i].time_went_active = 0;
		for(z=0;z<MAX_MODES;z++)
		{
			action_list[i].wcet[z] = 0;
		}
	}


	//************  Reset the timestamp iterators ************
	ce_block_i     = ce_blocks.begin();
	se_timestamp_i = se_timestamps.begin();
	rm_timestamp_i = rm_timestamps.begin();
	fr_timestamp_i = fr_timestamps.begin();
	md_timestamp_i = md_timestamps.begin();
	pe_timestamp_i = pe_timestamps.begin();




	//****************************************************************
	//** Process all the timestamps
	//****************************************************************
	while (get_next_timestamp_info(timestampInfo, TRUE)) 
	{
		zloops++;

		//*********************************************************
		//** add in windview information for the frame timestamps
		//*********************************************************
#ifndef __INTEGRATED_INTO_OTIF__
		if(bFrameTimestamps)
		{
			if((timestampInfo.strType == CE_TIMESTAMP) && (timestampInfo.bFrameController))
			{
				 if(timestampInfo.strAction == BEGIN_EVENT)
					{
//					 windview_frame(find_process_number(timestampInfo.pr), timestampInfo.llTimestamp);
					}
				   else
					{
//					 windview_frame_end(find_process_number(timestampInfo.pr),timestampInfo.llTimestamp);
//					 windview_idle(find_process_number(timestampInfo.pr),timestampInfo.llTimestamp);
					}
			}
			if(timestampInfo.bFrameController)
			{
				if(timestampInfo.strType == CE_TIMESTAMP) 
				{
					if(timestampInfo.iIndex >= 0)
					{
						 if (timestampInfo.strAction == BEGIN_EVENT)
							{
//								windview_process_entry(find_process_number(timestampInfo.pr),BEGIN, (unsigned __int64) timestampInfo.llTimestamp, framecontroller_subscribe_port_list[i].task_id, 55 /* tbd baa -- update thread priority */, (char*)timestampInfo.strSubscribePort.c_str() );
							}
						  else
							{
//								windview_process_entry(find_process_number(timestampInfo.pr),END, (unsigned __int64) timestampInfo.llTimestamp, framecontroller_subscribe_port_list[i].task_id, 55 /* tbd baa -- update thread priority */, (char*)timestampInfo.strSubscribePort.c_str()  );
							}
					}
				}
			}
		}
#endif


		//*********************************************************************
		//**  Update all of the known Subscribe Ports based on this timestamp
		//*********************************************************************
		for (i=0; i<subscribe_port_list_i; i++) 
		{
			strCompSubscribePort = subscribe_port_list[i].sp.name();

			switch (subscribe_port_list[i].nCurState) 
			{



				case Idle:

					if(subscribe_port_list[i].pr != timestampInfo.pr) break;

					if (!timestampInfo.bFrameController)
					{
						if (timestampInfo.strType == CE_TIMESTAMP)
						{
							if (timestampInfo.strSubscribePort == strCompSubscribePort)
							{
								if (!(timestampInfo.strAction == BEGIN_EVENT))
								{
									outStream << "Error: Event END out of order" << strCompSubscribePort << endl;
									cerr      << "Error: Event END out of order" << strCompSubscribePort << endl;
									if(!bIgnoreFatalErrors)
									{
									  _sleep(3000);
									  _exit(__LINE__);
									}
								}

								subscribe_port_list[i].nCurState       = Active;
								subscribe_port_list[i].recursion_level = 1;

								// Mark the absolute start time and the time component went active,
								// save the thread ID, and clear the running time sums.
								subscribe_port_list[i].llTimeStarted    = timestampInfo.llTimestamp ;
								subscribe_port_list[i].llTimeWentActive = timestampInfo.llTimestamp ;
								subscribe_port_list[i].strThreadID      = timestampInfo.strThreadID;
								subscribe_port_list[i].llExecutionTime  = 0;
								subscribe_port_list[i].llWallClockTime  = 0;
#ifndef __INTEGRATED_INTO_OTIF__
//								windview_process_entry(find_process_number(timestampInfo.pr),BEGIN, (unsigned __int64) timestampInfo.llTimestamp, subscribe_port_list[i].task_id, 55 /* tbd baa -- update thread priority */, (char*)timestampInfo.strSubscribePort.c_str()  );
#endif
								break;
							}
						}
					}
					break;


								
				
				
				case Active:					

					if(subscribe_port_list[i].pr != timestampInfo.pr) break;

					//****** If this is a consumed event START on the component's thread ********
					if ((timestampInfo.strType == CE_TIMESTAMP) &&
						(!timestampInfo.bFrameController) &&
						(timestampInfo.strAction == BEGIN_EVENT) &&
						(timestampInfo.strThreadID == subscribe_port_list[i].strThreadID))
					{
						// Same as the current component
						if(timestampInfo.strSubscribePort == strCompSubscribePort)
						{
							subscribe_port_list[i].recursion_level++;
							break;
						}

						// Ok, it is a different component
						// Change the component state to event preempted.
						subscribe_port_list[i].nCurState = EventPreempted;
						subscribe_port_list[i].llTimePreempted = timestampInfo.llTimestamp;

						// Calculate the intermediate time sum.
						subscribe_port_list[i].llExecutionTime +=
							timestampInfo.llTimestamp  - subscribe_port_list[i].llTimeWentActive;

						// Save the preemptor subscribe port. 
						subscribe_port_list[i].strPreemptor = timestampInfo.strSubscribePort;
					    break;
					}  
						

					//****** If this is a thread PREEMPTION start on the component's thread. ****
					if ((timestampInfo.strType == PE_TIMESTAMP) &&
						(timestampInfo.strAction == LEAVE_THREAD) &&
						(timestampInfo.strThreadID == subscribe_port_list[i].strThreadID)) 
					{

						// Change the component state to thread preempted.
						subscribe_port_list[i].nCurState = ThreadPreempted;

						// Calculate the intermediate time sum.
						subscribe_port_list[i].llExecutionTime +=
							timestampInfo.llTimestamp  - subscribe_port_list[i].llTimeWentActive;
						break;
					}
						
					//****** If this is a consumed event END for the component.*******************
					if ((timestampInfo.strType == CE_TIMESTAMP) &&
						(!timestampInfo.bFrameController) &&
						(timestampInfo.strAction == END_EVENT) &&
						(timestampInfo.strSubscribePort == strCompSubscribePort)) 
					{

						// check to see if this run is all done
						subscribe_port_list[i].recursion_level--;
						if(subscribe_port_list[i].recursion_level > 0)
						{
							break;
						}

						// Change the component state to idle.
						subscribe_port_list[i].nCurState = Idle;

						// Calculate the final execution time sum.
						subscribe_port_list[i].llExecutionTime +=
							timestampInfo.llTimestamp  - subscribe_port_list[i].llTimeWentActive;

						// Calculate the wall clock time sum.
						subscribe_port_list[i].llWallClockTime =
							timestampInfo.llTimestamp  - subscribe_port_list[i].llTimeStarted;

						int cm = get_component_current_mode( subscribe_port_list[i].group, subscribe_port_list[i].item);

						// Determine the worst case execution time.
						if (subscribe_port_list[i].llExecutionTime > subscribe_port_list[i].llWCExecutionTime[cm]) {
							subscribe_port_list[i].llWCExecutionTime[cm] = subscribe_port_list[i].llExecutionTime;
						}

						// Determine the worst case wall clock time.
						if (subscribe_port_list[i].llWallClockTime > subscribe_port_list[i].llWCWallClockTime[cm]) {
							subscribe_port_list[i].llWCWallClockTime[cm] = subscribe_port_list[i].llWallClockTime;
						}

						char strExecutionTime[128], strWallClockTime[128];
						sprintf(strExecutionTime, "%I64d", subscribe_port_list[i].llExecutionTime);
						sprintf(strWallClockTime, "%I64d", subscribe_port_list[i].llWallClockTime);

#ifndef __INTEGRATED_INTO_OTIF__
						if(output_timestamps)
						outStream << strCompSubscribePort.c_str() << " Exection time = " << strExecutionTime << endl;
						if(output_timestamps)
						outStream << strCompSubscribePort.c_str() << " Wall clock time = "<< strWallClockTime << endl;
//						windview_idle(  find_process_number(timestampInfo.pr),timestampInfo.llTimestamp );
#endif
						break;
					}

					break;



				case EventPreempted:

					if(subscribe_port_list[i].pr != timestampInfo.pr) break;

					// If this is a consumed event end on the same thread for the preemptor.
					if ((timestampInfo.strType == CE_TIMESTAMP) &&
						(!timestampInfo.bFrameController) &&
						(timestampInfo.strAction == END_EVENT) &&
						(timestampInfo.strThreadID == subscribe_port_list[i].strThreadID) &&
						(timestampInfo.strSubscribePort == subscribe_port_list[i].strPreemptor)) 
					{

						// Change the component state to active.
						subscribe_port_list[i].nCurState = Active;

						// Mark the time that the component went active.
						subscribe_port_list[i].llTimeWentActive = timestampInfo.llTimestamp ;
						subscribe_port_list[i].WCBT = timestampInfo.llTimestamp - subscribe_port_list[i].llTimePreempted;
					}
					break;



				case ThreadPreempted:

					if(subscribe_port_list[i].pr != timestampInfo.pr) break;

					// If this is thread preemption end on the component's thread.
					if ((timestampInfo.strType == PE_TIMESTAMP) &&
						(timestampInfo.strAction == RETURN_TO_THREAD) &&
						(timestampInfo.strThreadID == subscribe_port_list[i].strThreadID)) {

						// Change the component state to active.
						subscribe_port_list[i].nCurState = Active;

						// Mark the time that the component went active.
						subscribe_port_list[i].llTimeWentActive = timestampInfo.llTimestamp ;
					}
					break;



				default:
					outStream << "Error: Bad time stamp stream" << endl;
					cerr      << "Error: Bad time stamp stream" << endl;
					if(!bIgnoreFatalErrors)
					{
					  _sleep(3000);
					  _exit(__LINE__);
					}
					break;

			} // end switch on component state


		} // end for each subscribe port



		//*********************************************************************
		//**  Update all of the known actions based on this timestamp
		//*********************************************************************
		for (i=0; i<action_list_i; i++) 
		{
			string action_name;

			action_name = action_list[i].a.name();

			switch (action_list[i].nCurState) 
			{

				case Idle:
					// tbd baa -- if(!(action_list[i].pr       == timestampInfo.pr)) break;
					if(!(timestampInfo.strType   == RM_TIMESTAMP))     break;
					if(!(timestampInfo.strAction == BEGIN_RM))         break;
					if(!(timestampInfo.action    == action_list[i].a)) break;

					action_list[i].nCurState       = Active;

					// Mark the absolute start time and the time component went active,
					// save the thread ID, and clear the running time sums.
					action_list[i].time_went_active = timestampInfo.llTimestamp ;
					action_list[i].time_consumed    = 0;
					action_list[i].strThreadID      = timestampInfo.strThreadID;
					break;												
				
				case Active:					

					// tbd baa if(!(action_list[i].pr != timestampInfo.pr))   break;
					if(!(timestampInfo.strThreadID == action_list[i].strThreadID))  break;

					if(timestampInfo.strAction == LEAVE_THREAD)
					{
						if(!(timestampInfo.strType == PE_TIMESTAMP))   break;

						action_list[i].nCurState = ThreadPreempted;
						action_list[i].time_consumed += timestampInfo.llTimestamp  - action_list[i].time_went_active;						
						break;
					}

					if(timestampInfo.strAction == END_RM)
					{
						if(!(timestampInfo.strType == RM_TIMESTAMP))   break;

						action_list[i].nCurState = Idle;
						action_list[i].time_consumed += timestampInfo.llTimestamp  - action_list[i].time_went_active;

						string name;

						name = timestampInfo.action.name();

#ifndef __INTEGRATED_INTO_OTIF__
						if(output_timestamps)
						outStream << "Action " << name  <<  " execution time " << __i2s(action_list[i].time_consumed);
#endif

						int cm = get_component_current_mode(action_list[i].group, action_list[i].item );
						string mn = iif_get_component_mode_name(action_list[i].group, action_list[i].item ,cm);

						if(action_list[i].time_consumed > action_list[i].wcet[cm])
						{
#ifndef __INTEGRATED_INTO_OTIF__
						    if(output_timestamps)
							outStream << "  is now the WCET for this action under mode=" << mn;
#endif
							action_list[i].wcet[cm] = action_list[i].time_consumed;
						}
#ifndef __INTEGRATED_INTO_OTIF__
						if(output_timestamps)
						outStream << endl;
#endif
					}
					break;



				case EventPreempted:
					 action_list[i].nCurState = Idle;
					 break;


				case ThreadPreempted:

					// tbd baa -- if(action_list[i].pr != timestampInfo.pr) break;
 
					// If this is thread preemption end on the right thread.
					if (!(timestampInfo.strType == PE_TIMESTAMP))                  break;
					if(!(timestampInfo.strAction == RETURN_TO_THREAD))             break;
				    if(!(timestampInfo.strThreadID == action_list[i].strThreadID)) break;
					
						action_list[i].nCurState = Active;
						action_list[i].time_went_active = timestampInfo.llTimestamp ;
					break;



				default:
					outStream << "Error: Bad remote method stamp stream" << endl;
					cerr      << "Error: Bad remote method stamp stream" << endl;
					if(!bIgnoreFatalErrors)
					{
					  _sleep(3000);
					  _exit(__LINE__);
					}
					break;

			} // end switch on action state


		} // end for each action



		//**********************************************************************************
		//** Check for next timestamp is at the same time
		//**
        //** If the next timestamp is at the same time, we do not check for state errors
		//**   NOTICE the "continue" statement
		//**********************************************************************************
		TIMESTAMP_INFO peak_ahead;
		if (get_next_timestamp_info( peak_ahead, FALSE ))
		{
		 static unsigned __int64 last_timestamp = 0;

			if(peak_ahead.llTimestamp == last_timestamp) continue;

#ifndef __INTEGRATED_INTO_OTIF__
			if(peak_ahead.llTimestamp < last_timestamp)
			{

				outStream << "Error: Time has gone Noncausal was: " << __i2s(last_timestamp) << " now: " << __i2s(peak_ahead.llTimestamp) <<endl;
				cerr      << "Error: Time has gone Noncausal was: " << __i2s(last_timestamp) << " now: " << __i2s(peak_ahead.llTimestamp) <<endl;
				if(!bIgnoreFatalErrors)
				{
				  _sleep(3000);
				  _exit(__LINE__);
				}
			}
#endif

			last_timestamp = peak_ahead.llTimestamp;
		}

		//**********************************************************************************
		//** Check for concurrent subscribe port errors
		//**
        //** Check for two subscribe ports active at the same time.
		//** This doesn't work for windows version of data since no preemption time stamps.  
		//** Check is performed on a process only basis since things really can be concurrent
		//**********************************************************************************

		int p;
		for(p=0;p<process_list_i;p++)
		{
			bool bSubscribePortAlreadyActive = FALSE;
			int other_i;

			for (i=0; i<subscribe_port_list_i; i++) 
			{
                // only this process
				if(process_list[p] != subscribe_port_list[i].pr) continue;

				// If the current subscribe port is active.
				if (Active == subscribe_port_list[i].nCurState) 
				{

					// If another subscribe port is already active.
					if (bSubscribePortAlreadyActive) 
						{
							
							// Log the error and quit.
							char strTimeStamp[128];
							string  now;
							string then;

							now  = subscribe_port_list[i].sp.name();
							then = subscribe_port_list[other_i].sp.name();

							sprintf(strTimeStamp, "%I64d", timestampInfo.llTimestamp);
							outStream << "Error: " << now	<< " active  while "<< then << " active Timestamp = " << strTimeStamp << endl;
							cerr      << "Error: " << now	<< " active  while "<< then << " active Timestamp = " << strTimeStamp << endl;
							outStream << "    index=" << i << " other_i=" << other_i << endl;
							cerr      << "    index=" << i << " other_i=" << other_i << endl;
							if(!bIgnoreFatalErrors)
							{
							  _sleep(3000);
							  _exit(__LINE__);
							}
						} 
					  else 
						{
							// Record that a subscribe port is active.
							bSubscribePortAlreadyActive = TRUE;
							other_i = i;
						}
				}
			}
		}
	
    } // end while timestamp data 




	//************************************************************************
	//** Display the calculated times for WCET and wall clock and
	//** update the AIF with the new values.
	//************************************************************************
	
	for (i=0; i<subscribe_port_list_i; i++) 
	{
		char strWCExecutionTime[128], strWCWallClockTime[128];
		strCompSubscribePort = subscribe_port_list[i].sp.name();

		int z;

		for(z=0;z<MAX_MODES;z++)
		{
			string mn = iif_get_component_mode_name( subscribe_port_list[i].group, subscribe_port_list[i].item, z);
			if(mn == "") 
				break;
			if(z > 0)
				_sleep(1);
			sprintf(strWCExecutionTime, "%I64d", subscribe_port_list[i].llWCExecutionTime[z]);
			sprintf(strWCWallClockTime, "%I64d", subscribe_port_list[i].llWCWallClockTime[z]);

#ifndef __INTEGRATED_INTO_OTIF__
			outStream << endl;
			outStream << strCompSubscribePort.c_str() << " mode=" << mn << " Worst Case Exection time = "	<< strWCExecutionTime << endl;
			outStream << strCompSubscribePort.c_str() << " mode=" << mn << " Worst Case Wall clock time = " << strWCWallClockTime << endl;
#endif
		}
#ifndef __INTEGRATED_INTO_OTIF__
		outStream << strCompSubscribePort.c_str() << " Worst Case Blocking time = " << __i2s(subscribe_port_list[i].WCBT) << endl;
#endif

		if(subscribe_port_list[i].WCBT > largest_blocking_time) 
		{
			largest_blocking_time = subscribe_port_list[i].WCBT;
		}

		aif_set_wcet(subscribe_port_list[i]);
	}

#ifndef __INTEGRATED_INTO_OTIF__
	outStream << endl;
#endif

	for(i=0;i<action_list_i;i++)
	{
		int z;

		for(z=0;z<MAX_MODES;z++)
		{
			string mn = iif_get_component_mode_name( action_list[i].group, action_list[i].item, z);
			if(mn == "") 
				break;
			if(z > 0)
				_sleep(1);
#ifndef __INTEGRATED_INTO_OTIF__
			outStream << (string)(action_list[i].cp.name()) << " action=" << (string)(action_list[i].a.name()) << " mode=" << mn << " invocation wcet=" << __i2s(action_list[i].wcet[z]) << endl;
#endif
			aif_set_invocation_wcet( action_list[i], mn, z);
		}
	}

#ifndef __INTEGRATED_INTO_OTIF__
	outStream << endl << "Largest blocking time: " << __i2s(largest_blocking_time) << endl;
#endif
	return;
}




//****************************************************************************************
// Function initializes the IIF data object.
//****************************************************************************************
void iif_init(char* iif_file)
{
	iif = new Udm::SmartDataNetwork(diagram);

	try 
	{	
		if ((_access(iif_file, 0)) != -1) {

			// Open the input file and get the root object.
			iif->OpenExisting(iif_file,"", Udm::CHANGES_LOST_DEFAULT);
			iif_system = IIF::System::Cast(iif->GetRootObject());

			vector<IIFObject> iifobjects = iif_system.IIFObject_kind_children();
			vector<IIFObject>::iterator iifobjects_i = iifobjects.begin();
			int xx = 0;
			while (iifobjects_i != iifobjects.end())
			{
				string iifobject_name = iifobjects_i->name();
				outStream << "IIF object ";
				outStream << xx;
				outStream << " is ";
				outStream << iifobject_name;
				outStream<< endl;
				xx++;
				iifobjects_i++;
			}
		} else {

			outStream << "Error: IIF file " << iif_file << " not found." << endl;
			cerr << "Error: IIF file " << iif_file << " not found." << endl;
			_exit(__LINE__);
		}

	}

	catch(const udm_exception &e)
	{
		outStream << "Exception: initializing iif data: " << e.what() << endl;
		_sleep(4000);
		exit(__LINE__);

	} 

	outStream << endl;

	collect_component_instances();
	collect_framecontrollers();
	collect_threads();
    iif_show_component_instance_list();

#ifndef __INTEGRATED_INTO_OTIF__
//	windview_lock_idle();
#endif

	return;
}







//****************************************************************************************
// Function initializes and sorts the various timestamp sets.
//****************************************************************************************
void iif_init_timestamps(void)
{
	int nTimestamps = 0;
	int i;

	static 	set<ConsumedEventTimeStamp> zConsumedEventTimeStamps = iif_system.ConsumedEventTimeStamp_kind_children();
	set<ConsumedEventTimeStamp>::iterator zConsumedEventTimeStamps_i;

	for(i=0,zConsumedEventTimeStamps_i = zConsumedEventTimeStamps.begin(); zConsumedEventTimeStamps_i != zConsumedEventTimeStamps.end(); zConsumedEventTimeStamps_i++,i++)
	{
		unsigned __int64       this_time;
		ConsumedEventTimeStamp this_event;
		CE_TIMESTAMP_BLOCK this_block;


        this_time  = __s2u(zConsumedEventTimeStamps_i->timestamp());
		this_event = *zConsumedEventTimeStamps_i;

		this_block.ce   = this_event;
		this_block.time = this_time;

		ce_blocks.push_back( this_block );
	}
	
	se_timestamps = iif_system.SuppliedEventTimeStamp_kind_children();
	rm_timestamps = iif_system.RemoteMethodTimeStamp_kind_children();
	fr_timestamps = iif_system.FrameTimeStamp_kind_children();
	md_timestamps = iif_system.ModeTimeStamp_kind_children();
	pe_timestamps = iif_system.PreemptionTimeStamp_kind_children();


#ifndef __INTEGRATED_INTO_OTIF__
	outStream << "\nSorting ConsumedEvent Timestamps";
#endif
    sort(ce_blocks.begin(), ce_blocks.end());
#ifndef __INTEGRATED_INTO_OTIF__
	outStream << ".......done" << endl;
#endif

#ifndef __INTEGRATED_INTO_OTIF__
	outStream << "Sorting SuppliedEvent Timestamps";
#endif

	sort(se_timestamps.begin(), se_timestamps.end(), se_comp_timestamps);
#ifndef __INTEGRATED_INTO_OTIF__
	outStream << ".......done" << endl;
#endif

#ifndef __INTEGRATED_INTO_OTIF__
	outStream << "Sorting RemoteMethod Timestamps";
#endif

	sort(rm_timestamps.begin(), rm_timestamps.end(), rm_comp_timestamps);
#ifndef __INTEGRATED_INTO_OTIF__
	outStream << ".......done" << endl;
#endif

#ifndef __INTEGRATED_INTO_OTIF__
	outStream << "Sorting Frame Timestamps";
#endif

	sort(fr_timestamps.begin(), fr_timestamps.end(), fr_comp_timestamps);
#ifndef __INTEGRATED_INTO_OTIF__
	outStream << ".......done" << endl;
#endif

#ifndef __INTEGRATED_INTO_OTIF__
	outStream << "Sorting Mode Timestamps";
#endif

	sort(md_timestamps.begin(), md_timestamps.end(), md_comp_timestamps);
#ifndef __INTEGRATED_INTO_OTIF__
	outStream << ".......done" << endl;
#endif

#ifndef __INTEGRATED_INTO_OTIF__
	outStream << "Sorting Preemption Timestamps";
#endif

	sort(pe_timestamps.begin(), pe_timestamps.end(), pe_comp_timestamps);
#ifndef __INTEGRATED_INTO_OTIF__
	outStream << ".......done" << endl << endl;
#endif

	ce_block_i     = ce_blocks.begin();
	se_timestamp_i = se_timestamps.begin();
	rm_timestamp_i = rm_timestamps.begin();
	fr_timestamp_i = fr_timestamps.begin();
	md_timestamp_i = md_timestamps.begin();
	pe_timestamp_i = pe_timestamps.begin();
}


//*********************************************************************************************

void insert_preemption( int old_thread_index, int new_thread_index, unsigned __int64 timevalue )
{
  // tbd baa -- would want to insert them here
}


//****************************************************************************************
// Function processes the timestamp lists, getting the associated IIF information.
//****************************************************************************************
void iif_process_timestamps(IIF_PROCESS_TIMESTAMP_PASS_OPTION pass_option)
{
	int nTimestamps = 0;
	TIMESTAMP_TYPE c;
	char szAction[10];
	int x,tx;
	bool frame_controller_event;
	int active_thread = -1;

	// Reset the timestamp iterators
	ce_block_i     = ce_blocks.begin();
	se_timestamp_i = se_timestamps.begin();
	rm_timestamp_i = rm_timestamps.begin();
	fr_timestamp_i = fr_timestamps.begin();
	md_timestamp_i = md_timestamps.begin();
	pe_timestamp_i = pe_timestamps.begin();
	
	for(;;nTimestamps++)
	{
		c = next_timestamp_type();
		switch(c)
		{
			case eConsumedEvent:
			{
				ConsumedEventTimeStamp this_ce_timestamp = ce_block_i->ce;
				SubscribePort          ce_subscribe_port = this_ce_timestamp.SubscribePort();
				Thread                 ce_thread = this_ce_timestamp.thread();

				//****************************************************************************************************************
				//** Print the trace information to the text file
				//****************************************************************************************************************
				string strtname = ce_thread.name();
				char tname[512];
				string frame_controller_sp_name;
				sprintf(tname, "%-20s", strtname.c_str());

				if (this_ce_timestamp.type() & 1) {
					sprintf(szAction, " %s    ", END_EVENT);
				} else {
					sprintf(szAction, " %s  ", BEGIN_EVENT);
				}

#ifndef __INTEGRATED_INTO_OTIF__
				if(output_timestamps)
				outStream << CE_TIMESTAMP << " " << __i2s(ce_block_i->time) <<	szAction << tname;
#endif

				frame_controller_event = FALSE;
				if ((x = find_matching_component_instance( iif_system, ce_subscribe_port )) < 0)
					{
						if ((x = find_matching_framecontroller( iif_system, ce_subscribe_port )) >= 0)
						{
#ifndef __INTEGRATED_INTO_OTIF__
				            if(output_timestamps)
								outStream << " " <<  (string)framecontroller_subscribe_port_list[x].sp.name();
#endif
							frame_controller_event = TRUE;
							frame_controller_sp_name = framecontroller_subscribe_port_list[x].sp.name();
						}
					}
#ifndef __INTEGRATED_INTO_OTIF__
				   else
					{
				        if(output_timestamps)
						outStream << " " <<  (string)subscribe_port_list[x].sp.name();
					}
				if(output_timestamps)
					outStream << endl;
#endif

				//****************************************************************************************************************
				// update thread state based on fake preemptions if needed 
				//****************************************************************************************************************
				tx = find_thread( ce_thread );
				if((tx != active_thread)  && (pass_option == CREATE_FAKE_PREEMPTIONS))
				{
					insert_preemption( active_thread, tx, ce_block_i->time );
					active_thread = tx;
				}


				//****************************************************************************************************************
				//** Create fake time stamps based on frame controller
				//**
				//** Early versions of the IIF data files did not support the beginning of the frame time stamps.
				//** As such, the frame start was derived by watching the frame controller component (which does
				//** not exist in the UML models and seeing when it actually started.  This time is used as the frame start
				//****************************************************************************************************************
				if (((this_ce_timestamp.type() & 1) == 0)  && frame_controller_event)
				{
					if(tx >= 0)
					{
						if(thread_list[tx].infrastructure_overhead != overhead_frame_start) 
						{
						  if((thread_list[tx].infrastructure_overhead != overhead_frame_controller) || (frame_controller_sp_name == "UU_DFS_timeoutConsumer" /* allow reset of frame start on new timeout */))
						  {
							  thread_list[tx].infrastructure_start_time = ce_block_i->time;
							  bWarningDetected = TRUE;

#ifndef __INTEGRATED_INTO_OTIF__
							  outStream << "Warning: #### missing frame start detected -- synthesized " << __i2s(ce_block_i->time) << "   " << tname<< endl;
#endif

							  thread_list[tx].infrastructure_overhead = overhead_frame_controller;
							
							  thread_list[tx].time_resumed              = ce_block_i->time;
							  thread_list[tx].unaccounted_for_time      = 0;
							  thread_list[tx].state                     = active;
						  }
						}
					}
				}

				//****************************************************************************************************************
				//** Roll in infrastructure overhead
				//**
				//** If enabled, roll in infrastructure overhead at each component begin time
				//** Accomplished by simply changing the timestamp value and letting the next pass use the revised time
				//****************************************************************************************************************
				if (((this_ce_timestamp.type() & 1) == 0)   && (!frame_controller_event)  && (thread_list[tx].infrastructure_start_time != 0) && (thread_list[tx].infrastructure_overhead != unknown) && (x >= 0))
				{
					if(tx >= 0)
					{
						if(pass_option == ROLL_IN_INFRASTRUCTURE)
						{
						 unsigned __int64 tsv;

						  tsv = ce_block_i->time;
                          ce_block_i->time = thread_list[tx].infrastructure_start_time; 
#ifndef __INTEGRATED_INTO_OTIF__
				          if(output_timestamps)
						  outStream << CE_TIMESTAMP << " " << __i2s(ce_block_i->time) <<
											szAction << tname << " " <<  (string)subscribe_port_list[x].sp.name() << " ### REVISED to roll in infrastructure ###" << endl;
#endif

						}
						thread_list[tx].infrastructure_start_time = 0;
						thread_list[tx].infrastructure_overhead = overhead_unknown;
					}
				}

				//****************************************************************************************************************
				//** Mark/save frame start time as consume event time for component subscribe port for a particular thread
				//** This allows chains of components in the same rate group to overlap infrastrucutre overhead
				//** That is, the next component in the chain can be considered to start when the previous one finished.
				//** This time will be used when ROLL_IN_INFRASTRUCTURE is used
				//****************************************************************************************************************
				if (((this_ce_timestamp.type() & 1) != 0)   && (!frame_controller_event)  && (thread_list[tx].infrastructure_start_time == 0))
				{
					if(tx >= 0)
					{
						thread_list[tx].infrastructure_start_time = ce_block_i->time;
						thread_list[tx].infrastructure_overhead   = overhead_previous_component;
					}
				}
				ce_block_i++;
			}
			break;


			case eSuppliedEvent:
			{
				SuppliedEventTimeStamp se_timestamp = *se_timestamp_i;
				if (se_timestamp.type() & 1) {
					sprintf(szAction, " %s    ", END_EVENT);
				} else {
					sprintf(szAction, " %s  ", BEGIN_EVENT);
				}
				char tname[512];
				sprintf(tname, "%-20s", NO_DATA_FIELD);

#ifndef __INTEGRATED_INTO_OTIF__
				if(output_timestamps)
					outStream << SE_TIMESTAMP << " " << __i2s(se_timestamp.timestamp()) << szAction << tname;
#endif

				Thread se_thread = se_timestamp.thread();
				PublishPort se_publish_port = se_timestamp.PublishPort();

				tx = find_thread( se_thread );
				if((tx != active_thread)  && (pass_option == CREATE_FAKE_PREEMPTIONS))
				{
					insert_preemption( active_thread, tx, __s2u(se_timestamp.timestamp()) );
					active_thread = tx;
				}

#ifndef __INTEGRATED_INTO_OTIF__
				if((x = find_matching_component_instance( iif_system, se_publish_port )) < 0)
				{
					if ((x = find_matching_framecontroller( iif_system, se_publish_port )) >= 0)
					{
				        if(output_timestamps)
						outStream << " " <<  (string)framecontroller_publish_port_list[x].pp.name();
					}
				}
				else
				{
					if(output_timestamps)
					outStream << " " <<  (string)publish_port_list[x].pp.name();
				}

				if(output_timestamps)
					outStream << endl;
#endif
				se_timestamp_i++;
			}
			break;

			case eFrame:
			{
				FrameTimeStamp fr_timestamp = *fr_timestamp_i;

				if (fr_timestamp.type() & 1) {
					sprintf(szAction, " %s    ", END_EVENT);
				} else {
					sprintf(szAction, " %s  ", BEGIN_EVENT);
				}

#ifndef __INTEGRATED_INTO_OTIF__
				if(output_timestamps)
					outStream << FR_TIMESTAMP << " " << __i2s(fr_timestamp.timestamp()) << szAction;
#endif

				Rate rate = fr_timestamp.Rate();
				x = find_rate(rate);

				char tname[512];
				if (x >= 0) {
					string strtname = thread_list[x].definition.name();
					sprintf(tname, "%-20s", strtname.c_str());

#ifndef __INTEGRATED_INTO_OTIF__
					if(output_timestamps)
						outStream << tname;
#endif

					thread_list[x].infrastructure_start_time = __s2u(fr_timestamp.timestamp());
					thread_list[x].infrastructure_overhead   = overhead_frame_start;
				} else {
					sprintf(tname, "%-20s", NO_DATA_FIELD);

#ifndef __INTEGRATED_INTO_OTIF__
				    if(output_timestamps)
						outStream << tname ;
#endif

				}

				string rname = rate.name();

#ifndef __INTEGRATED_INTO_OTIF__
				if(output_timestamps)
					outStream << " " << rname << endl;				
#endif

				fr_timestamp_i++;
			}
			break;

			case eMode:
			{
				ModeTimeStamp md_timestamp = *md_timestamp_i;
				string sthis_mode;

				sthis_mode = update_component_instance_modes( iif_system, md_timestamp );

#ifndef __INTEGRATED_INTO_OTIF__
				if(output_timestamps)
					outStream << MD_TIMESTAMP << " " << __i2s(md_timestamp.timestamp()) << " " << sthis_mode << endl;
#endif

				md_timestamp_i++;
			}
			break;

			case eRemoteMethod:
			{
				RemoteMethodTimeStamp rm_timestamp = *rm_timestamp_i;
				Action action = rm_timestamp.Action();
				Thread th = rm_timestamp.thread();
				string tname = th.name();
				string aname = action.name();
				string atype;

				switch(rm_timestamp.type())
				{
				case 2: atype = "CBegin"; break;
				case 3: atype = "CEnd  "; break;
				case 0: atype = "SBegin"; break;
				case 1: atype = "SEnd  "; break;
				}

#ifndef __INTEGRATED_INTO_OTIF__
				if(output_timestamps)
					outStream << "rm_timestamp ";
				if(output_timestamps)
					outStream <<  __i2s(rm_timestamp.timestamp()) << " " << atype << " " << tname << "\t\t\t  " << aname << endl; 
#endif

				rm_timestamp_i++;
			}
			break;

			case ePreemption:
			{
				PreemptionTimeStamp pe_timestamp = *pe_timestamp_i;

				if (pe_timestamp.type()) {
					sprintf(szAction, " %s ", RETURN_TO_THREAD);
				} else {
					sprintf(szAction, " %s  ", LEAVE_THREAD);
				}

#ifndef __INTEGRATED_INTO_OTIF__
				if(output_timestamps)
					outStream << PE_TIMESTAMP << " " << __i2s(pe_timestamp.timestamp()) << szAction;
#endif

				Thread th = pe_timestamp.Thread();
				char tname[512];
				if ((x = find_thread(th)) >= 0) {
					string strtname = thread_list[x].definition.name();
					sprintf(tname, "%-20s", strtname.c_str());

#ifndef __INTEGRATED_INTO_OTIF__
					if(output_timestamps)
						outStream << tname;
#endif

					active_thread = x;
				} else {
					sprintf(tname, "%-20s", NO_DATA_FIELD);

#ifndef __INTEGRATED_INTO_OTIF__
					if(output_timestamps)
						outStream << tname ;
#endif

					active_thread = -1;
				}

#ifndef __INTEGRATED_INTO_OTIF__
				if(output_timestamps)
					outStream << " " << NO_DATA_FIELD << endl;
#endif

				pe_timestamp_i++;
			}
			break;

			case eNone:
			{
				if(pass_option == ROLL_IN_INFRASTRUCTURE)
				{

#ifndef __INTEGRATED_INTO_OTIF__
					outStream << "\nResorting ConsumedEvent Timestamps -- since they have been corrected to hold infrastructure overhead";
#endif

					sort(ce_blocks.begin(), ce_blocks.end());

#ifndef __INTEGRATED_INTO_OTIF__
					outStream << ".......done" << endl;
#endif

				}
				if(pass_option == CREATE_FAKE_PREEMPTIONS)
				{

#ifndef __INTEGRATED_INTO_OTIF__
					outStream << "\nResorting Preemption Timestamps -- since they have been corrected";
#endif

					sort(pe_timestamps.begin(), pe_timestamps.end(), pe_comp_timestamps);

#ifndef __INTEGRATED_INTO_OTIF__
					outStream << ".......done" << endl;
#endif

				}
				return;
			}
			break;

		} // end switch()

	} // end for()
}















